@@ -14,83 +14,30 @@
 # See the License for the specific language governing permissions and limitations under the License.
 
 """
-Pure Python implementation of CEL.
-
-This provides a few jq-like, bc-like, and shell expr-like features.
-
--   ``jq`` uses ``.`` to refer the current document. By setting a package
-    name of ``"jq"`` and placing the JSON object in the package, we achieve
-    similar syntax.
-
--   ``bc`` offers complex function definitions and other programming support.
-    CEL can only evaluate a few bc-like expressions.
-
--   This does everything ``expr`` does, but the syntax is slightly different.
-    The output of comparisons -- by default -- is boolean, where ``expr`` is an integer 1 or 0.
-    Use ``-f 'd'`` to see decimal output instead of Boolean text values.
-
--   This does some of what ``test`` does, without a lot of the sophisticated
-    file system data gathering.
-    Use ``-b`` to set the exit status code from a Boolean result.
-
-TODO: This can also have a REPL, as well as process CSV files.
-
-SYNOPSIS
-========
-
-::
-
-    python -m celpy [--arg name:type=value ...] [--null-input] expr
-
-Options:
-
-:--arg:
-    Provides argument names, types and optional values.
-    If the value is not provided, the name is expected to be an environment
-    variable, and the value of the environment variable is converted and used.
-
-:--null-input:
-    Normally, JSON documents are read from stdin in ndjson format. If no JSON documents are
-    provided, the ``--null-input`` option skips trying to read from stdin.
-
-:expr:
-    A CEL expression to evaluate.
-
-JSON documents are read from stdin in NDJSON format (http://jsonlines.org/, http://ndjson.org/).
-For each JSON document, the expression is evaluated with the document in a default
-package. This allows `.name` to pick items from the document.
-
-By default, the output is JSON serialized. This means strings will be JSON-ified and have quotes.
-
-If a ``--format`` option is provided, this is applied to the resulting object; this can be
-used to strip quotes, or limit precision on double objects, or convert numbers to hexadecimal.
-
-Arguments, Types, and Namespaces
-================================
-
-CEL objects rely on the celtypes definitions.
-
-Because of the close association between CEL and protobuf, some well-known protobuf types
-are also supported.
-
-..  todo:: CLI type environment
-
-    Permit name.name:type=value to create namespace bindings.
-
-Further, type providers can be bound to CEL. This means an extended CEL
-may have additional types beyond those defined by the :py:class:`Activation` class.
+The CLI interface to ``celpy``.
 
+This parses the command-line options.
+It also offers an interactive REPL.
 """
 
 import argparse
 import ast
 import cmd
+import datetime
 import json
 import logging
+import logging.config
 import os
+from pathlib import Path
 import re
+import stat as os_stat
 import sys
-from typing import Any, Callable, Dict, List, Optional, Tuple, cast
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast
+
+try:
+    import tomllib
+except ImportError:  # pragma: no cover
+    import tomli as tomllib  # type: ignore [no-redef]
 
 from celpy import Environment, Runner, celtypes
 from celpy.adapter import CELJSONDecoder, CELJSONEncoder
@@ -104,29 +51,23 @@ logger = logging.getLogger("celpy")
 # Note the reliance on `ast.literal_eval` for ListType and MapType conversions.
 # Other types convert strings directly. These types need some help.
 CLI_ARG_TYPES: Dict[str, Annotation] = {
-    "int":
-        celtypes.IntType,
-    "uint":
-        celtypes.UintType,
-    "double":
-        celtypes.DoubleType,
-    "bool":
-        celtypes.BoolType,
-    "string":
-        celtypes.StringType,
-    "bytes":
-        celtypes.BytesType,
-    "list":
-        cast(Callable[..., celtypes.Value], lambda arg: celtypes.ListType(ast.literal_eval(arg))),
-    "map":
-        cast(Callable[..., celtypes.Value], lambda arg: celtypes.MapType(ast.literal_eval(arg))),
-    "null_type":
-        cast(Callable[..., celtypes.Value], lambda arg: None),
-    "single_duration":
-        celtypes.DurationType,
-    "single_timestamp":
-        celtypes.TimestampType,
-
+    "int": celtypes.IntType,
+    "uint": celtypes.UintType,
+    "double": celtypes.DoubleType,
+    "bool": celtypes.BoolType,
+    "string": celtypes.StringType,
+    "bytes": celtypes.BytesType,
+    "list": cast(
+        Callable[..., celtypes.Value],
+        lambda arg: celtypes.ListType(ast.literal_eval(arg)),
+    ),
+    "map": cast(
+        Callable[..., celtypes.Value],
+        lambda arg: celtypes.MapType(ast.literal_eval(arg)),
+    ),
+    "null_type": cast(Callable[..., celtypes.Value], lambda arg: None),
+    "single_duration": celtypes.DurationType,
+    "single_timestamp": celtypes.TimestampType,
     "int64_value": celtypes.IntType,
     "uint64_value": celtypes.UintType,
     "double_value": celtypes.DoubleType,
@@ -175,7 +116,8 @@ def arg_type_value(text: str) -> Tuple[str, Annotation, celtypes.Value]:
     match = arg_pattern.match(text)
     if match is None:
         raise argparse.ArgumentTypeError(
-            f"arg {text} not 'var=string', 'var:type=value', or `var:type")
+            f"arg {text} not 'var=string', 'var:type=value', or `var:type"
+        )
     name, type_name, value_text = match.groups()
     if value_text is None:
         value_text = os.environ.get(name)
@@ -186,12 +128,16 @@ def arg_type_value(text: str) -> Tuple[str, Annotation, celtypes.Value]:
             type_definition = CLI_ARG_TYPES[type_name]
             value = cast(
                 celtypes.Value,
-                type_definition(value_text)  # type: ignore[arg-type, call-arg]
+                type_definition(value_text),  # type: ignore [call-arg]
             )
         except KeyError:
-            raise argparse.ArgumentTypeError(f"arg {text} type name not in {list(CLI_ARG_TYPES)}")
+            raise argparse.ArgumentTypeError(
+                f"arg {text} type name not in {list(CLI_ARG_TYPES)}"
+            )
         except ValueError:
-            raise argparse.ArgumentTypeError(f"arg {text} value invalid for the supplied type")
+            raise argparse.ArgumentTypeError(
+                f"arg {text} value invalid for the supplied type"
+            )
     else:
         value = celtypes.StringType(value_text or "")
         type_definition = celtypes.StringType
@@ -199,56 +145,83 @@ def arg_type_value(text: str) -> Tuple[str, Annotation, celtypes.Value]:
 
 
 def get_options(argv: Optional[List[str]] = None) -> argparse.Namespace:
-    """Parses command-line arguments.
-    """
+    """Parses command-line arguments."""
     parser = argparse.ArgumentParser(prog="celpy", description="Pure Python CEL")
-    parser.add_argument(
-        "-v", "--verbose", default=0, action='count')
+    parser.add_argument("-v", "--verbose", default=0, action="count")
 
     # Inputs
     parser.add_argument(
-        "-a", "--arg", action='append', type=arg_type_value,
+        "-a",
+        "--arg",
+        action="append",
+        type=arg_type_value,
         help="Variables to set; -a name:type=value, or -a name=value for strings, "
-             "or -a name to read an environment variable"
+        "or -a name to read an environment variable",
     )
     parser.add_argument(
-        "-n", "--null-input", dest='null_input', default=False, action='store_true',
-        help="Avoid reading Newline-Delimited JSON documents from stdin"
+        "-n",
+        "--null-input",
+        dest="null_input",
+        default=False,
+        action="store_true",
+        help="Avoid reading Newline-Delimited JSON documents from stdin",
     )
     parser.add_argument(
-        "-s", "--slurp", default=False, action="store_true",
-        help="Slurp a single, multiple JSON document from stdin"
+        "-s",
+        "--slurp",
+        default=False,
+        action="store_true",
+        help="Slurp a single, multiple JSON document from stdin",
     )
     parser.add_argument(
-        "-i", "--interactive", default=False, action="store_true",
-        help="Interactive REPL"
+        "-i",
+        "--interactive",
+        default=False,
+        action="store_true",
+        help="Interactive REPL",
     )
 
     # JSON handling
     parser.add_argument(
-        "--json-package", "-p", metavar="NAME", dest="package", default=None, action="store",
-        help="Each JSON input is a CEL package, allowing .name to work"
+        "--json-package",
+        "-p",
+        metavar="NAME",
+        dest="package",
+        default=None,
+        action="store",
+        help="Each JSON input is a CEL package, allowing .name to work",
     )
     parser.add_argument(
-        "--json-document", "-d", metavar="NAME", dest="document", default=None, action="store",
-        help="Each JSON input is a variable, allowing name.map(x, x*2) to work"
+        "--json-document",
+        "-d",
+        metavar="NAME",
+        dest="document",
+        default=None,
+        action="store",
+        help="Each JSON input is a variable, allowing name.map(x, x*2) to work",
     )
 
     # Outputs and Status
     parser.add_argument(
-        "-b", "--boolean", default=False, action='store_true',
-        help="If the result is True, the exit status is 0, for False, it's 1, otherwise 2"
+        "-b",
+        "--boolean",
+        default=False,
+        action="store_true",
+        help="If the result is True, the exit status is 0, for False, it's 1, otherwise 2",
     )
     parser.add_argument(
-        "-f", "--format", default=None, action='store',
+        "-f",
+        "--format",
+        default=None,
+        action="store",
         help=(
             "Use Python formating instead of JSON conversion of results; "
             "Example '.6f' to format a DoubleType result"
-        )
+        ),
     )
 
     # The expression
-    parser.add_argument("expr", nargs='?')
+    parser.add_argument("expr", nargs="?")
 
     options = parser.parse_args(argv)
     if options.package and options.document:
@@ -262,10 +235,77 @@ def get_options(argv: Optional[List[str]] = None) -> argparse.Namespace:
     return options
 
 
+def stat(path: Union[Path, str]) -> Optional[celtypes.MapType]:
+    """This function is added to the CLI to permit file-system interrogation."""
+    try:
+        status = Path(path).stat()
+        data = {
+            "st_atime": celtypes.TimestampType(
+                datetime.datetime.fromtimestamp(status.st_atime)
+            ),
+            "st_ctime": celtypes.TimestampType(
+                datetime.datetime.fromtimestamp(status.st_ctime)
+            ),
+            "st_mtime": celtypes.TimestampType(
+                datetime.datetime.fromtimestamp(status.st_mtime)
+            ),
+            "st_dev": celtypes.IntType(status.st_dev),
+            "st_ino": celtypes.IntType(status.st_ino),
+            "st_nlink": celtypes.IntType(status.st_nlink),
+            "st_size": celtypes.IntType(status.st_size),
+            "group_access": celtypes.BoolType(status.st_gid == os.getegid()),
+            "user_access": celtypes.BoolType(status.st_uid == os.geteuid()),
+        }
+
+        # From mode File type:
+        # - block, character, directory, regular, symbolic link, named pipe, socket
+        # One predicate should be True; we want the code for that key.
+        data["kind"] = celtypes.StringType(
+            {
+                predicate(status.st_mode): code
+                for code, predicate in [
+                    ("b", os_stat.S_ISBLK),
+                    ("c", os_stat.S_ISCHR),
+                    ("d", os_stat.S_ISDIR),
+                    ("f", os_stat.S_ISREG),
+                    ("p", os_stat.S_ISFIFO),
+                    ("l", os_stat.S_ISLNK),
+                    ("s", os_stat.S_ISSOCK),
+                ]
+            }.get(True, "?")
+        )
+
+        # Special bits: uid, gid, sticky
+        data["setuid"] = celtypes.BoolType((os_stat.S_ISUID & status.st_mode) != 0)
+        data["setgid"] = celtypes.BoolType((os_stat.S_ISGID & status.st_mode) != 0)
+        data["sticky"] = celtypes.BoolType((os_stat.S_ISVTX & status.st_mode) != 0)
+
+        # permissions, limited to user-level RWX, nothing more.
+        data["r"] = celtypes.BoolType(os.access(path, os.R_OK))
+        data["w"] = celtypes.BoolType(os.access(path, os.W_OK))
+        data["x"] = celtypes.BoolType(os.access(path, os.X_OK))
+        try:
+            extra = {
+                "st_birthtime": celtypes.TimestampType(
+                    datetime.datetime.fromtimestamp(status.st_birthtime)
+                ),
+                "st_blksize": celtypes.IntType(status.st_blksize),
+                "st_blocks": celtypes.IntType(status.st_blocks),
+                "st_flags": celtypes.IntType(status.st_flags),
+                "st_rdev": celtypes.IntType(status.st_rdev),
+                "st_gen": celtypes.IntType(status.st_gen),
+            }
+        except AttributeError:  # pragma: no cover
+            extra = {}
+        return celtypes.MapType(data | extra)
+    except FileNotFoundError:
+        return None
+
+
 class CEL_REPL(cmd.Cmd):
     prompt = "CEL> "
     intro = "Enter an expression to have it evaluated."
-    logger = logging.getLogger("REPL")
+    logger = logging.getLogger("celpy.repl")
 
     def cel_eval(self, text: str) -> celtypes.Value:
         try:
@@ -273,7 +313,10 @@ class CEL_REPL(cmd.Cmd):
             prgm = self.env.program(expr)
             return prgm.evaluate(self.state)
         except CELParseError as ex:
-            print(self.env.cel_parser.error_text(ex.args[0], ex.line, ex.column), file=sys.stderr)
+            print(
+                self.env.cel_parser.error_text(ex.args[0], ex.line, ex.column),
+                file=sys.stderr,
+            )
             raise
 
     def preloop(self) -> None:
@@ -285,7 +328,7 @@ class CEL_REPL(cmd.Cmd):
 
         Evaluates the expression, saves the result as the given variable in the current activation.
         """
-        name, space, args = args.partition(' ')
+        name, space, args = args.partition(" ")
         try:
             value: celtypes.Value = self.cel_eval(args)
             print(value)
@@ -305,6 +348,7 @@ class CEL_REPL(cmd.Cmd):
 
     do_exit = do_quit
     do_bye = do_quit
+    do_EOF = do_quit
 
     def default(self, args: str) -> None:
         """Evaluate an expression."""
@@ -316,12 +360,13 @@ class CEL_REPL(cmd.Cmd):
 
 
 def process_json_doc(
-        display: Callable[[Result], None],
-        prgm: Runner,
-        activation: Dict[str, Any],
-        variable: str,
-        document: str,
-        boolean_to_status: bool = False) -> int:
+    display: Callable[[Result], None],
+    prgm: Runner,
+    activation: Dict[str, Any],
+    variable: str,
+    document: str,
+    boolean_to_status: bool = False,
+) -> int:
     """
     Process a single JSON document. Either one line of an NDJSON stream
     or the only document in slurp mode. We assign it to the variable "jq".
@@ -332,10 +377,10 @@ def process_json_doc(
     """
     try:
         activation[variable] = json.loads(document, cls=CELJSONDecoder)
-        result = prgm.evaluate(activation)
-        display(result)
-        if (boolean_to_status and isinstance(result, (celtypes.BoolType, bool))):
-            return 0 if result else 1
+        result_value = prgm.evaluate(activation)
+        display(result_value)
+        if boolean_to_status and isinstance(result_value, (celtypes.BoolType, bool)):
+            return 0 if result_value else 1
         return 0
     except CELEvalError as ex:
         # ``jq`` KeyError problems result in ``None``.
@@ -378,11 +423,13 @@ def main(argv: Optional[List[str]] = None) -> int:
         return 0
 
     if options.format:
-        def output_display(result: Result) -> None:
-            print('{0:{format}}'.format(result, format=options.format))
+
+        def output_display(result_value: Result) -> None:
+            print("{0:{format}}".format(result_value, format=options.format))
     else:
-        def output_display(result: Result) -> None:
-            print(json.dumps(result, cls=CELJSONEncoder))
+
+        def output_display(result_value: Result) -> None:
+            print(json.dumps(result_value, cls=CELJSONEncoder))
 
     logger.info("Expr: %r", options.expr)
 
@@ -391,55 +438,65 @@ def main(argv: Optional[List[str]] = None) -> int:
 
     annotations: Optional[Dict[str, Annotation]]
     if options.arg:
-        annotations = {
-            name: type for name, type, value in options.arg
-        }
+        annotations = {name: type for name, type, value in options.arg}
     else:
-        annotations = None
+        annotations = {}
+    annotations["stat"] = celtypes.FunctionType
 
     # If we're creating a named JSON document, we don't provide a default package.
-    # If we're usinga  JSON document to populate a package, we provide the given name.
+    # If we're using a  JSON document to populate a package, we provide the given name.
     env = Environment(
         package=None if options.null_input else options.package,
         annotations=annotations,
     )
     try:
         expr = env.compile(options.expr)
-        prgm = env.program(expr)
+        prgm = env.program(expr, functions={"stat": stat})
     except CELParseError as ex:
-        print(env.cel_parser.error_text(ex.args[0], ex.line, ex.column), file=sys.stderr)
+        print(
+            env.cel_parser.error_text(ex.args[0], ex.line, ex.column), file=sys.stderr
+        )
         return 1
 
     if options.arg:
-        activation = {
-            name: value for name, type, value in options.arg
-        }
+        activation = {name: value for name, type, value in options.arg}
     else:
         activation = {}
 
     if options.null_input:
         # Don't read stdin, evaluate with only the activation context.
         try:
-            result = prgm.evaluate(activation)
+            result_value = prgm.evaluate(activation)
             if options.boolean:
-                if isinstance(result, (celtypes.BoolType, bool)):
-                    summary = 0 if result else 1
+                if isinstance(result_value, (celtypes.BoolType, bool)):
+                    summary = 0 if result_value else 1
                 else:
-                    logger.warning("Expected celtypes.BoolType, got %s = %r", type(result), result)
+                    logger.warning(
+                        "Expected celtypes.BoolType, got %s = %r",
+                        type(result_value),
+                        result_value,
+                    )
                     summary = 2
             else:
-                output_display(result)
+                output_display(result_value)
                 summary = 0
         except CELEvalError as ex:
-            print(env.cel_parser.error_text(ex.args[0], ex.line, ex.column), file=sys.stderr)
+            print(
+                env.cel_parser.error_text(ex.args[0], ex.line, ex.column),
+                file=sys.stderr,
+            )
             summary = 2
 
     elif options.slurp:
         # If slurp, one big document, part of the "jq" package in the activation context.
         document = sys.stdin.read()
         summary = process_json_doc(
-            output_display, prgm, activation, options.document or options.package, document,
-            options.boolean
+            output_display,
+            prgm,
+            activation,
+            options.document or options.package,
+            document,
+            options.boolean,
         )
 
     else:
@@ -450,16 +507,43 @@ def main(argv: Optional[List[str]] = None) -> int:
             summary = max(
                 summary,
                 process_json_doc(
-                    output_display, prgm, activation, options.document or options.package, document,
-                    options.boolean
-                )
+                    output_display,
+                    prgm,
+                    activation,
+                    options.document or options.package,
+                    document,
+                    options.boolean,
+                ),
             )
 
     return summary
 
 
+CONFIG_PATHS = (dir_path / "celpy.toml" for dir_path in (Path.cwd(), Path.home()))
+
+DEFAULT_CONFIG_TOML = """
+[logging]
+  version = 1
+  formatters.minimal.format = "%(message)s"
+  formatters.console.format = "%(levelname)s:%(name)s:%(message)s"
+  formatters.details.format = "%(levelname)s:%(name)s:%(module)s:%(lineno)d:%(message)s"
+  root.level = "WARNING"
+  root.handlers = ["console"]
+
+[logging.handlers.console]
+    class = "logging.StreamHandler"
+    formatter = "console"
+"""
+
+
 if __name__ == "__main__":  # pragma: no cover
-    logging.basicConfig(level=logging.WARNING)
-    exit = main(sys.argv[1:])
+    config_paths = list(p for p in CONFIG_PATHS if p.exists())
+    config_toml = config_paths[0].read_text() if config_paths else DEFAULT_CONFIG_TOML
+    log_config = tomllib.loads(config_toml)
+    if "logging" in log_config:
+        logging.config.dictConfig(log_config["logging"])
+
+    exit_status = main(sys.argv[1:])
+
     logging.shutdown()
-    sys.exit(exit)
+    sys.exit(exit_status)
